/*
	Blizzard - Heroes of the Storm - M3 importer
	
	Written by Taylor Mouse
*/

/* GLOBALS */
global mdlName = ""
global allBones 		= #()
global allSeqs 			= #()
global allIndices 		= #()
global allVertices 		= #()
global allNormals 		= #()
global allBoneWeights 	= #()
global allBoneIndices 	= #()
global allUVWs 			= #()
global allAttachments 	= #()
global allSubMeshes 	= #()
global allSTCs			= #()

fn ResetGlobals=
(
	 mdlName = ""
	 allBones 		= #()
	 allSeqs 			= #()
	 allIndices 		= #()
	 allVertices 		= #()
	 allNormals 		= #()
	 allBoneWeights 	= #()
	 allBoneIndices 	= #()
	 allUVWs 			= #()
	 allAttachments 	= #()
	 allSubMeshes 	= #()
	
)


/*
	The most important part is the tag !
*/
struct SC2Tag 	  ( Name, OffSet, Length = 0, Flags )
struct SC2Bone 	  ( Id, Name,  ParentBoneId, BonePos, BoneRot, BoneScale, BoneType, TheBone )
struct SC2Seqs	  ( Id, Name ) 
struct SC2Att	  ( Id, Name, ParentId )
struct SC2Submesh ( Id, Name, firstVertex, nVerts, firstIndex, nIndices, TheMesh )
struct SC2STC  	  ( Name, SequenceInd, AnimId, AnimRef, AnimOffset, SequenceData, CollectionStart, CollectionEnd )


/**************************************************************************************************************************/
/*  HELPER FUNCTIONS */
/**************************************************************************************************************************/
/*
	Build layer
*/
fn CreateLayer layerName =
(
	layer = LayerManager.GetLayerFromName layerName
	if(layer == undefined) then 
	(
		layer = LayerManager.NewLayerFromName layerName
	)
	return layer
)
/*
	Read a specific number of bytes, without returning anything
*/
fn ReadBytes stream val =
(	
	for i=1 to val do ReadByte stream
)
/* 
	Read four bytes in reverse and return the string value
*/
fn ReadFourCC stream =
(
	ret = ""
	
	for i=1 to 4 do
	(
		ret = (bit.IntAsChar(ReadByte stream)) + ret
	)
	
	return ret
)
/* 
	Read a string defined with a fixed length
*/
fn ReadFixedString stream val =
(
	local str = ""
	for i=1 to val do
	(
		s= bit.IntAsChar(ReadByte stream)
		
		if(s!="\0") then str+=s
	)
	return str
)

/**************************************************************************************************************************/
/*  READ THE M3 */
/**************************************************************************************************************************/
fn ReadM3 file =
(
	-- file = @"F:\StarCraft II Exported\Heroes Of The Storm\Extracted\Models\Storm_Building_KingsCrest_Core_Dragon.m3"
		
	stream = fOpen file "rb"
	id = ReadFourCC stream

	if( id != "MD33" and id != "MD34" ) then  return
	
	ofTag = 0
	ofTag 	= ReadLong stream #unsigned 
	nTag 	= ReadLong stream #unsigned
	entries = ReadLong stream #unsigned
	refId 	= ReadLong stream #unsigned
	
	
	/* READ THE TAGS */
	tags = #()
	
	fSeek stream ofTag #seek_set
	
	for i=1 to nTag do
	(
		local tag 	= SC2Tag()
		tag.Name 	= ReadFourCC stream
		tag.Offset 	= ReadLong stream
		tag.Length 	= ReadLong stream
		tag.Flags 	= ReadLong stream
		
		append tags tag
	)
	
	/* LOOP THROUGH THE TAGS */
	for i=1 to nTag do
	(
		local tag = tags[i]
	
		/* MODEL NAME */
		if(tag.Name == "MODL") then
		(
			fSeek stream tags[i+1].OffSet #seek_set
			
			mdlName = GetFilenameFile ( ReadFixedString stream  tags[i+1].Length )
		) --// END MODL
		
		/* BONES */
		if(tag.Name == "BONE" ) then
		(
			/* BONE NAMES */
			for b=1 to tag.Length do
			(
				local aBone = SC2Bone()
				
				if( tags[i+b].Name == "CHAR") then
				(
					fSeek stream tags[i+b].OffSet #seek_set
					aBone.Name = ReadFixedString stream  tags[i+b].Length
					aBone.Id = b
				)
				
				append allBones aBone
			)
			
			
			/* BONE DATA 
			
			*/
			/* BONE TYPES
				0  - Root or other
				32 - Hit Test
				34 - Dummy or IK Chain
				40 - Connection Bone
				42 - Bone
			*/
			
			fSeek stream tag.OffSet #seek_set
		
			size = ( tags[i+1].Offset - tag.OffSet ) / tag.Length
			
			-- print size --> 160
				
			for b=1 to tag.Length do
			(
				ReadBytes stream 4 					--> FF FF FF FF
				nameLen = ReadLong stream #unsigned --> Bone name length -> covered in the reading of the CHAR data
				id = ReadLong stream #unsigned
				unk1 = ReadLong stream #unsigned 	--> 00 00 00 00 
				unk2 = ReadByte stream
				
				boneType = ReadShort stream #signed -- Bone Type
				unk3 = ReadByte stream
				-- 20 bytes
				
				local parentBoneId = (ReadShort stream #signed ) + 1 -- Parent Bone --> if boneId = 0 then it has no parent

				unk4 = ReadBytes stream 2
				unk5 = ReadShort stream #signed
				unk6 = ReadShort stream #signed
				unk7 = ReadLong stream #unsigned
				-- 32 bytes
				
				bonePos = [(ReadFloat stream) * 50.0, (ReadFloat stream) * 50.0, (ReadFloat stream) * 50.0 ]
				-- 44
				unk8 = ReadBytes stream 24
				-- 68
				boneRot = quat (ReadFloat stream) (ReadFloat stream) (ReadFloat stream) (ReadFloat stream)
				-- 84
				unk9 = ReadBytes stream 28
				--112
				boneScale = [(ReadFloat stream) , (ReadFloat stream) , (ReadFloat stream)]
				-- 124 bytes
				
				ReadBytes stream (size - 124)
				
				/*
				SC2Bone 	  
				( Id, Name, ParentBoneId, BonePos, BoneRot, BoneScale, BoneType, BoneRef )
				*/
				aBone = allBones[b]
				aBone.ParentBoneId = parentBoneId
				aBone.BoneType = boneType
				aBone.BonePos = bonePos
				aBone.BoneRot = boneRot
				aBone.BoneScale = boneScale
			
			)
			--print allBones
			
		) --// END BONES
		
		if(tag.Name =="SEQS" ) then
		(
			for s=1 to tag.Length do
			(
				local seq = SC2Seqs()			
				fSeek stream tags[i + s ].OffSet #seek_set
				seq.Id = s
				seq.Name = ReadFixedString stream  tags[i + s].Length
				
				append allSeqs seq
			)
		) -- // END SEQS
		
		if(tag.Name =="DIV_" ) then --// Indices
		(
			fSeek stream tag.OffSet #seek_set
			
			nIndices = ReadShort stream #unsigned
			size  = ReadLong  stream #unsigned
			
			for s=1 to tag.Length do
			(
				fSeek stream tags[i+s].OffSet #seek_set
				
				for t=1 to tags[i+s].Length do
				(
					-- f = [(( ReadShort stream #unsigned ) +1 ), (( ReadShort stream #unsigned ) +1 ), (( ReadShort stream #unsigned ) +1 ) ]
					append allIndices (( ReadShort stream #unsigned ) +1 )
				)

			)
		) -- // END INDICES
		
		if(tag.Name =="U8__" ) then --// Vertices
		(
			fSeek stream 128 #seek_set
		
			flag = 0
			chunkSize = 0
			flag = ReadLong stream #unsigned
			
			if(flag == 25297021) then chunkSize = 32
			if(flag == 25297533) then chunkSize = 36
			if(flag == 25296993) then chunkSize = 32
			if(flag == 25559649) then chunkSize = 40
			if(flag == 25297505) then chunkSize = 36
			if(flag == 25559677) then chunkSize = 40
			if(flag == 25559137) then chunkSize = 36
			if(flag == 25559165) then chunkSize = 36
			if(flag == 26083937) then chunkSize = 44
			if(flag == 26083453) then chunkSize = 40
			if(flag == 26083425) then chunkSize = 40
			if(flag == 27132001) then chunkSize = 44
			if(flag == 27132513) then chunkSize = 48
			if(flag == 1099038817) then chunkSize = 36
			if(flag == 1099300961) then chunkSize = 40
			if(flag == 1099038845) then chunkSize = 36
			if(flag == 1099300989) then chunkSize = 40
			if(flag == 27132029) then chunkSize = 44
			
			nVerts = tag.Length / chunkSize
			
			fSeek stream tag.OffSet #seek_set
			
			for v=1 to nVerts do
			(
				x1 = ReadFloat stream * 50
				y1 = ReadFloat stream * 50
				z1 = ReadFloat stream * 50
				
				append allVertices [ x1, y1, z1 ] 
				
				x2 = ((Readbyte stream #unsigned) as float) / 255.0 
				y2 = ((Readbyte stream #unsigned) as float) / 255.0 
				z2 = ((Readbyte stream #unsigned) as float) / 255.0 
				w2 = ((Readbyte stream #unsigned) as float) / 255.0 
				
				append allBoneWeights [ x2, y2, z2, w2 ] -- weight refers to the bone in the allBoneIndices
				
				x3 = (ReadByte stream #unsigned) +1
				y3 = (ReadByte stream #unsigned) +1
				z3 = (ReadByte stream #unsigned) +1
				w3 = (ReadByte stream #unsigned) +1
							
				append allBoneIndices 	[ x3, y3, z3, w3 ] -- max 4 bones per index
			
				x4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				y4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				z4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				w4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1

				append allNormals 		[ x4, y4, z4]
				
				uvCorrection = 2048.0
				u5 = (ReadShort stream #signed ) / uvCorrection
				v5 = 1.0 - ((ReadShort stream #signed ) / uvCorrection )
				w5 = 0.0 
				append allUVWs [ u5,v5,w5]
				
				ReadBytes stream (chunkSize - 28)
			)
		) -- // END VERTICES
		
		if(tag.Name == "ATT_" ) then -- // Attachment points
		(
			for b=1 to tag.Length do
			(
				local att = SC2Att()
				
				if( tags[i+b].Name == "CHAR") then
				(
					fSeek stream tags[i+b].OffSet #seek_set
					att.Id = b
					att.Name = ReadFixedString stream  tags[i+b].Length
				)
				append allAttachments att
			)
			
			size = (tags[i+1].OffSet - tag.OffSet) / tag.Length 
			print size
			fSeek stream tag.OffSet #seek_set
			
			for b=1 to tag.Length do
			(
				ReadBytes stream 16
				allAttachments[b].parentID = (ReadLong stream #unsigned) + 1
				
			)
			
		) --// END ATTACHMENT POINTS
		
		if(tag.Name == "REGN" ) then -- // Submeshes
		(
			fSeek stream tag.OffSet #seek_set
			
			for s=1 to tag.Length do
			(
				local submesh = SC2Submesh()
				
				ReadBytes stream 4
				
				if( id != "MD33") then
				(
					ReadBytes stream 4
					submesh.firstVertex = (ReadLong stream #unsigned) as integer
					submesh.nVerts = (ReadLong stream #unsigned) as integer
				)
				else
				(
					submesh.firstVertex = (ReadShort stream #unsigned) as integer
					submesh.nVerts = (ReadShort stream #unsigned) as integer
				)
				submesh.Id = s
				submesh.Name = "Mesh " + (s as string)
				submesh.firstIndex = (ReadLong stream #unsigned) as integer
				submesh.nIndices = (ReadLong stream #unsigned) as integer
				
				ReadBytes stream 12
				
				append allSubMeshes submesh
			)
		) --// END SUBMESHES
		
		
		
	) -- // END TAGS

	fClose stream
)

/**************************************************************************************************************************/
/*  BUILD THE MESHES */
/**************************************************************************************************************************/

fn BuildMeshes =
(
	nMeshes = allSubMeshes.count
	nFirstFace = 0
	
	/* Create the meshes */
	for m=1 to nMeshes do
	(
		submesh = allSubMeshes[m]
		--print submesh
		
		local v = #()
		local n = #()
		local uv = #()

		for x=1 to submesh.nVerts do
		(
			append v allVertices[submesh.firstVertex + x]
			append n allNormals[submesh.firstVertex + x]
			append uv allUVWs[submesh.firstVertex + x]
		)

		local f = #()
		
		for x=1 to submesh.nIndices by 3 do
		(
			local face = [0,0,0]
			
			for j=1 to 3 do
			(
				local k = submesh.firstIndex + x + j - 1
				face[j] = allIndices[k]
			)
			append f face
		)
					
		nFirstFace = lastx
		
		local theMesh = mesh vertices:v vnorms:n name:submesh.Name tverts:uv faces:f
		
		unitLayer = CreateLayer "Unit"
		theMesh.WireColor = Color (random 0 128) (random 0 255) (random 0 128)
		
		buildTVFaces theMesh false
		for i = 1 to f.count do
		( setTVFace theMesh i f[i] )
		
		update theMesh
		unitLayer.AddNode theMesh
		allSubMeshes[m].TheMesh = theMesh
	)
	
	redrawviews()
	
)
/**************************************************************************************************************************/
/*  BUILD THE BONES */
/**************************************************************************************************************************/

fn BuildBones = 
(
	nBones = allBones.count
	/* Put the bones in a layer */
	boneLayer = CreateLayer "Bones"

	
	/* Create the bones */
	for b=1 to nBones do
	(
		aBone = BoneSys.CreateBone [0,0,0] [0,0,0] [0,0,1]
		aBone.Name = allBones[b].Name
		aBone.ShowLinks = true
		aBone.Width = 0.05
		aBone.Height = 0.05
		
		allBones[b].TheBone = aBone
		
		boneLayer.AddNode aBone
	)
		
	/* Build the bone Hierarchy */
	for b=1 to nBones do
	(
		if(allBones[b].ParentBoneId != 0) then 
		(
			parentBone = allBones[ allBones[b].ParentBoneId ].TheBone
			allBones[b].TheBone.Parent = parentBone
		)
	)
		
	/* Build Skin Pose for each bone */
	in coordsys parent
	(
		for b=1 to nbones do 
		(
			aBone = allBones[b].TheBone
			aBone.Rotation = allBones[b].BoneRot 
			aBone.Position = allBones[b].BonePos
			aBone.Scale = allBones[b].BoneScale
			aBone.SetSkinPose()
		)
	)
	
	
	
)

/**************************************************************************************************************************/
/*  BUILD THE SKIN */
/**************************************************************************************************************************/

fn BuildSkin = 
(
	with redraw off
	
	nSubMeshes = allSubMeshes.count
	nBones = allBones.count
	
	for s=1 to nSubMeshes do
	(
		-- get a reference to the mesh
		msh = allSubMeshes[s].TheMesh
		
		-- apply skin modifier
		select msh
		max modify mode 
		modPanel.addModToSelection(skin())
		skinMod = msh.Modifiers["skin"]
		modPanel.setCurrentObject skinMod
		
		-- add the bones to the skin modifier
		for b=1 to nBones do 
		( 
			skinOps.addBone skinMod allBones[b].TheBone 0
		)
	
		-- apply the vertex weights
		nVertices = msh.vertices.count
	
		for v=1 to nVertices do
		(
			skinOps.ReplaceVertexWeights skinMod v allBoneIndices[v] allBoneWeights[v]
		)
		update msh
	)
	
)
/**************************************************************************************************************************/
/*  BUILD THE ATTACHMENTS */
/**************************************************************************************************************************/

fn BuildAttachments = 
(
	refLayer = CreateLayer "Refs"
	for i=1 to allAttachments.count do
	(
		boneId = allAttachments[i].parentID
		theBone = allBones[boneId].TheBone
		
		att = SC2AttachPt name:allAttachments[i].Name Scale:[0.5,0.5,0.5] pos:theBone.Position rot:theBone.Rotation
		att.parent = theBone
		--print att
		--d = Dummy name:allAttachments[i].Name Scale:[0.5,0.5,0.5] pos:theBone.Position rot:theBone.Rotation
		--d.parent = theBone
		refLayer.AddNode att
	)
)
/**************************************************************************************************************************/
/*  CALL THE FUNCTIONS */
/**************************************************************************************************************************/
/*
for obj in objects do delete obj
ClearListener()
	ResetGlobals()
	ReadM3 ""
	BuildMeshes()
	BuildBones()
	--BuildSkin()
	--BuildAttachments()
print allSeqs
*/










/************************************************************************************************
 UTILITY
************************************************************************************************/
Utility HeroesOfTheStorm "HotS - Model Import"
(
	Group "Heroes of the Storm"
	(
		Label lbl1 "(m3) model import"
		Label lbl2 "written by Taylor Mouse"
		Label lbl3 "2014"
		
		EditText txtSelectedFile
		Button bSelectFile "..."
		
		Button bGo "Import..." height:35 width:100
		
	)
			
	On bSelectFile PRESSED do
	(
		file = getOpenFileName \ 
			caption:"Heroes Of The Storm m3 File" \
			types:"Heroes Of The Storm model(*.m3)|*.m3" \
			historyCategory:"HotSObjectPresets"
		if(file!=undefined ) then txtSelectedFile.Text = file
	)
	On bGo PRESSED do
	(
		if(txtSelectedFile.Text != undefined and (DoesFileExist txtSelectedFile.Text )) then
		(
			ClearListener()
			ResetGlobals()
			ReadM3 txtSelectedFile.Text
			BuildMeshes()
			BuildBones()
			BuildSkin()
			BuildAttachments()
		)
	)
)
